package polyglot.ext.ibex.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.types.*;
import polyglot.ext.ibex.types.*;
import polyglot.ext.ibex.visit.*;
import polyglot.ext.ibex.lr.GLR;
import polyglot.visit.*;
import polyglot.util.*;
import java.util.*;

/**
 * Node representing a semantic action.
 */
public class TerminalDecl_c extends Term_c implements TerminalDecl {
    Flags flags;
    TypeNode type;
    String name;
    Terminal terminal;

    public TerminalDecl_c(Position pos, Flags flags, TypeNode type, String name) {
        super(pos);
        this.flags = flags;
        this.type = type;
        this.name = name;
    }

    public Flags flags() { return flags; }
    public TerminalDecl flags(Flags flags) {
        TerminalDecl_c n = (TerminalDecl_c) copy();
        n.flags = flags;
        return n;
    }

    public TypeNode type() { return type; }
    public TerminalDecl type(TypeNode type) {
        TerminalDecl_c n = (TerminalDecl_c) copy();
        n.type = type;
        return n;
    }

    public String name() { return name; }
    public TerminalDecl name(String name) {
        TerminalDecl_c n = (TerminalDecl_c) copy();
        n.name = name;
        return n;
    }
    
    public Terminal terminal() { return terminal; }
    public TerminalDecl terminal(Terminal terminal) {
        TerminalDecl_c n = (TerminalDecl_c) copy();
        n.terminal = terminal;
        return n;
    }

    protected TerminalDecl_c reconstruct(TypeNode type) {
        if (type != this.type) {
            TerminalDecl_c n = (TerminalDecl_c) copy();
            n.type = type;
            return n;
        }
        return this;
    }

    public Node visitChildren(NodeVisitor v) {
        TypeNode type = (TypeNode) visitChild(this.type, v);
        return reconstruct(type);
    }

    public Node buildTypes(TypeBuilder tb) throws SemanticException {
        IbexTypeSystem ts = (IbexTypeSystem) tb.typeSystem();

        ParserType ct = (ParserType) tb.currentClass();

        Terminal terminal = ts.terminal(position(), ct,
            Flags.NONE, ts.unknownType(position()), name);

        return terminal(terminal);
    }

    /** Build type objects for the method. */
    public NodeVisitor disambiguateEnter(AmbiguityRemover ar) throws SemanticException {
        if (ar.kind() == AmbiguityRemover.SUPER) {
            return ar.bypassChildren(this);
        }

        return ar;
    }

    public Node disambiguate(AmbiguityRemover ar) throws SemanticException {
        if (ar.kind() == AmbiguityRemover.SIGNATURES) {
            Context c = ar.context();
            IbexTypeSystem ts = (IbexTypeSystem) ar.typeSystem();

            MutableParserType ct = (MutableParserType) c.currentClassScope();

            Terminal terminal = makeTerminal(ct, ts);

            return terminal(terminal).flags(terminal.flags());
        }

        return this;
    }

    public NodeVisitor addMembersEnter(AddMemberVisitor am) {
        MutableParserType ct = (MutableParserType)
            am.context().currentClassScope();
        ct.addTerminal(terminal);
        ct.addField(terminal);
        return am.bypassChildren(this);
    }

    /** Type check the method. */
    public Node typeCheck(TypeChecker tc) throws SemanticException {
        IbexTypeSystem ts = (IbexTypeSystem) tc.typeSystem();

        if (! type.type().isSubtype(ts.runtimeTerminalType())) {
            throw new SemanticException("Terminal must implement interface \"" +
                                        ts.runtimeTerminalType().fullName() +
                                        "\".");
        }

        ParserType pt = (ParserType) tc.context().currentClass();

        if (pt.flags().isInterface()) {
            if (flags().isProtected() || flags().isPrivate()) {
                throw new SemanticException("Interface members must be public.",
                                            position());
            }
        }

        /*
        if (flags().isStatic()) {
            throw new SemanticException("Terminals cannot be static.", position());
        }
        */
        if (flags().isNative()) {
            throw new SemanticException("Terminals cannot be native.", position());
        }
        if (flags().isAbstract()) {
            throw new SemanticException("Terminals cannot be abstract.", position());
        }
        if (flags().isTransient()) {
            throw new SemanticException("Terminals cannot be transient.", position());
        }
        if (flags().isStrictFP()) {
            throw new SemanticException("Terminals cannot be strictfp.", position());
        }
        if (flags().isVolatile()) {
            throw new SemanticException("Terminals cannot be volatile.", position());
        }

        return this;
    }

    protected Terminal makeTerminal(ParserType ct, IbexTypeSystem ts)
        throws SemanticException {

        Flags flags = this.flags;

        // Terminals are implicitly final static.
        flags = flags.Static().Final();

        // Terminals in interfaces are implicitly public.
        if (ct.flags().isInterface()) {
            flags = flags.Public();
        }

        return ts.terminal(position(), ct, flags, type.type(), name);
    }

    /**
     * Return the first (sub)term performed when evaluating this
     * term.
     */
    public Term entry() {
        return this;
    }

    /**
     * Visit this term in evaluation order.
     */
    public List acceptCFG(CFGBuilder v, List succs) {
        return succs;
    }

    public List rewrite(Rewriter rw) throws SemanticException {
        NodeFactory nf = rw.nodeFactory();
        TypeSystem ts = rw.typeSystem();
        ParsedClassType ct = rw.context().currentClassScope();

        /*
        FieldInstance fi = ts.fieldInstance(
            position(), ct, flags, ts.Int(), name);
            */

        GLR glr = rw.glr();

        int symbolNum = glr.terminalNumber(terminal);

        FieldDecl f = 
            nf.FieldDecl(position(), flags,
                         nf.CanonicalTypeNode(position(), ts.Int()), name,
                         nf.IntLit(position(), IntLit.INT, symbolNum));
        f = f.fieldInstance(terminal);

        /*
        ct.addField(fi);
        */

        return Collections.singletonList(f);
    }
}
