package polyglot.ext.polyj.ast;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;
import polyglot.frontend.*;
import polyglot.frontend.Compiler;

import polyglot.ext.jl.ast.*;
import polyglot.ext.polyj.extension.*;
import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.visit.*;

import java.util.*;


public class PolyJClassBody_c extends ClassBody_c implements PolyJClassBody
{
    public PolyJClassBody_c( Position pos, List members) {
        super(pos, members);
    }
    
    public void translate(CodeWriter w, Translator tr) {
        super.translate(w, tr);
        
        PolyJTypeSystem ts = (PolyJTypeSystem)tr.typeSystem();
        Context c = tr.context();
        ClassType curClass = c.currentClass();
        ClassType curScope = c.currentClassScope();
        Flags flags = curClass.flags();
        boolean isInterface = flags.isInterface();
        String adIntName = PolyJTranslator.adapterInterfaceShortName(curClass);
        
        w.begin(4);
        w.newline();
        
        if (!isInterface
	    && ts.needsWhereObject(curClass) 
	    && !ts.inheritsWhereObject(curClass)) 
	{
            w.write( "protected " + adIntName + " $W;" );
            w.newline();
        }

        if (ts.needsWhereObject(curClass)) {
            insertTypeCheckingMethods(w, (PolyJTranslator) tr, isInterface, 
                curScope, adIntName);
        }

        if(!isInterface) {
            implementInterfaceTypeCheckingMethods(w, (PolyJTranslator) tr,
                adIntName, curScope);
        }
        
        if (((PolyJTranslator)tr).clinitNeeded(curClass)) {
            writeClinit(w, tr, curClass);
        }
        
        w.end();
        w.newline();
    }

    /**
     * Return an iterator over the members of the class body
     * that are affected by the translation of static initialization.
     */
    protected Iterator staticTransMembers() {
        return new TransformingIterator(members.iterator(), 
            new Transformation() {
                public Object transform(Object o) {
                    if (o instanceof FieldDecl) {
                        FieldDecl f = (FieldDecl) o;
                        if (f.flags().isStatic()) {
                            return f;
                        } else {
                            return Transformation.NOTHING;
                        }
                    } else if (o instanceof Initializer) {
                        return o;
                    } else {
                        return Transformation.NOTHING;
                    }
                }
            });
    }
    
    protected boolean clinitNeeded(Translator tr, ClassType curClass) {
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
        boolean isParametric = ts.needsWhereObject(curClass);
        boolean isInterface = curClass.flags().isInterface();
        boolean isInnerNonStatic = curClass.isInner() 
            && !curClass.flags().isStatic();
        return isParametric && !isInterface && !isInnerNonStatic;
    }

    protected void writeClinit(CodeWriter w, Translator tr, ClassType curClass) {
        String adIntName = PolyJTranslator.adapterInterfaceShortName(curClass);
        w.write("public static void $clinit(" + adIntName + " $W) { ");

        if (staticTransMembers().hasNext()) {
            w.newline(4);
            w.begin(0);
            for (Iterator i = staticTransMembers(); i.hasNext(); ) {
                ClassMember m = (ClassMember) i.next();

                if (m instanceof FieldDecl) {
                    FieldDecl f = (FieldDecl) m;
                    writeFieldInitCall(w, tr, f);
                } else if (m instanceof Initializer) {
                    Initializer init = (Initializer) m;
                    writeBlockInitCall(w, tr, init);
                }
                if (!i.hasNext()) {
                    w.end();
                }
                w.newline();
            }
        }
        
        w.write("}");
        w.newline(0);
    }
    
    protected void writeFieldInitCall(CodeWriter w, Translator tr, FieldDecl f) 
    {
        w.write("$init$" + f.name() + "($W);");
    }
    
    protected void writeBlockInitCall(CodeWriter w, Translator tr, 
        Initializer init) 
    {
        int initNum = ((PolyJInitializerExt) init.del()).initNum();
        w.write("$init$" + initNum + "($W);");
    }
    
    protected void insertTypeCheckingMethods(CodeWriter w, PolyJTranslator tr,
        boolean isInterface, ClassType curScope, String adIntName) 
    {
	boolean writeBody = !isInterface;
	
        // Insert the $cast$$Name method for this type.
        insertCast(w, tr, writeBody, curScope);

        // Insert the $instanceof$$Name method for this type.
        insertInstanceof(w, tr, writeBody, curScope, adIntName);

        w.newline(0);
    }

    protected void insertInstanceof(CodeWriter w, PolyJTranslator tr, 
	boolean writeBody, ClassType curScope, String adIntName) 
    {
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
        String fullName = PolyJTranslator.fullNameDollar(curScope);

        List allParams = ts.allTypeParams(curScope);
        int nParams = allParams.size();
        if (nParams < 1) {
            w.write("// No $instanceof$$" + fullName + " method needed; "
			    + "no type variables are in scope.");
            w.newline(0);
            return;
        }

        w.write("public boolean $instanceof$$" + fullName + "(String $1");
        for (int count=2; count <= nParams; count++) {
            w.write(", String $" + count);
        }
        w.write(" )");
	
	if (writeBody) {
	    w.write(" {");
            w.newline(4);
   	    w.begin(0);
	    w.write("return ");
	    for (int count=1; count <= nParams; count++) {
		if (count != 1) {
	    	    w.newline(4);
		    w.write("&& ");
		}
		ParamType param = (ParamType) allParams.get(count-1);
		String methodName = ts.ParamSigBinding(param, curScope).name();
		w.write("$" + count + ".equals( ((" + adIntName + ")" 
		    + tr.whereObjExpr() + ")." + methodName + "() )");
	    }
            w.write(";");
            w.end();
            w.newline(0);
            w.write("}");
            w.newline(0);
	} else {
	    w.write(";");
	    w.newline(0);
	}
    }

    protected void insertCast(CodeWriter w, Translator tr, 
	boolean writeBody, ClassType curScope) 
    {
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
        Context c = tr.context();

        String fullName = PolyJTranslator.fullNameDollar(curScope);

        List allParams = ts.allTypeParams(curScope);
        int nParams = allParams.size();
        if (nParams < 1) {
            w.write("// No $cast$$" + fullName + " method needed; "
			    + "no type variables are in scope.");
            w.newline(0);
            return;
        }

        w.write("public " + curScope.translate(c) + " $cast$$"
            + fullName + "(String $1");
        for (int count = 2; count <= nParams; count++) {
            w.write(", String $" + count);
        }
        w.write(")"); 
	
	if (writeBody) {
	    w.write(" {");
            w.newline(4);

            w.begin(0);
            w.write("if (! $instanceof$$" + fullName + "($1");
            for (int count=2; count <= nParams; count++) {
                w.write(", $" + count);
            }
            w.write(")) { throw new ClassCastException(); }");
            w.newline(0);
            w.write("return this;");
            w.end();
            w.newline(0);
            w.write( "}" );
            w.newline(0);	    
	} else {
	    w.write(";");
	    w.newline(0);
	}
    }

    protected void implementInterfaceTypeCheckingMethods(CodeWriter w,
	PolyJTranslator tr, String adIntName, ClassType curScope) 
    {
        // superInterfaces : Map[PolyJClassType,Set[InstType]]
        // giving the instantiations of every polymorphic type appearing
        // as a superinterface.
        Map superInterfaces = new HashMap();
	fetchPolySuperInterfaces(tr, curScope, superInterfaces);
	//System.out.println(type + " <: " + superInterfaces);
	
	Iterator i = superInterfaces.entrySet().iterator();
	while (i.hasNext()) {
            Map.Entry entry = (Map.Entry) i.next();
            PolyJClassType polyType = (PolyJClassType) entry.getKey();
            Set instantiations = (Set) entry.getValue();
            insertInstanceofInterface(w, tr, polyType, instantiations,
                adIntName, curScope);
        }
        w.newline(0);

        i = superInterfaces.keySet().iterator();
        while (i.hasNext()) {
	    PolyJClassType polyType = (PolyJClassType) i.next();
            insertCastToInterface(w, tr, polyType);
        }
	w.newline(0);
    }
    
    protected void insertInstanceofInterface(CodeWriter w, PolyJTranslator tr,
        PolyJClassType polyType, Set instantiations, String adIntName,
        ClassType curScope) 
    {
        String fullName = PolyJTranslator.fullNameDollar(polyType);
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
        List allParams = ts.allTypeParams(polyType);
        int nParams = allParams.size();
        
        w.write("public boolean $instanceof$$" + fullName + "(String $1" );
        for (int count=2; count <= nParams; count++) {
            w.write(", String $" + count);
        }
        w.write(" ) {");
        w.newline(4);
        w.begin(4);
        w.write("return ");
        
        Iterator i = instantiations.iterator();
        while (i.hasNext()) {
            ClassType interfaceInstantiation = (ClassType) i.next();
            List allActuals = ts.allTypeArgs(interfaceInstantiation);
	    String name = interfaceInstantiation.name();
            
            w.begin(0);
            w.write( "// is this a " + name + " ?");
            w.newline(0);
            w.write("(   $1.equals( " 
	        + getParamString(tr, (Type) allActuals.get(0), 
                    adIntName, curScope) + " )");
            for (int count=1; count < nParams; count++) {
                w.newline(1);
                w.write( "&& $"+ (count+1) +".equals( "
                    + getParamString(tr, (Type) allActuals.get(count),
                        adIntName, curScope) + " )" );
            }        
            w.write( " )" );
            w.end();

            if (i.hasNext()) { 
                w.newline(0); 
                w.write( "|| ");
            } else {
                w.write( ";" );
                w.end();
                w.newline(0);
            }
        }
        
        w.write( "}" );
        w.newline(0);
    }

    protected void insertCastToInterface(CodeWriter w, Translator tr,
        PolyJClassType polyType) 
    {
	String interfaceName = PolyJTranslator.fullNameForType(polyType);
        String fullName = PolyJTranslator.fullNameDollar(polyType);
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
        List allParams = ts.allTypeParams(polyType);
        int nParams = allParams.size();

        w.write("public " + interfaceName + " $cast$$" + fullName 
	    + "( String $1" );
        for (int count=2; count <= nParams; count++) {
            w.write(", String $" + count);
        }
        w.write(" ) {");
        w.newline(4);
        w.begin(0);
        w.write("if (! $instanceof$$" + fullName + "($1");
        for (int count=2; count <= nParams; count++) {
            w.write( ", $"+count );
        }
        w.write(")) { throw new ClassCastException(); }");
        w.newline(0);
        w.write("return this;"); 
        w.end();
        w.newline(0);
        w.write("}"); 
        w.newline(0);
    }

    
    /**
     * Finds a string representation of param to use for instanceof 
     * checking.
     * If param is a parameter of this class (e.g. the class implements "A[T]"),
     *  then this function returns a PSig method.
     * If param is not a parameter of this class (e.g. the class implemenents
     *  "A[int]"), then this function returns the appropriate type descriptor.
     */
    protected String getParamString(PolyJTranslator tr, Type interfaceParam,
        String adIntName, ClassType curScope) 
    {
        if (interfaceParam instanceof ParamType) {
            ParamType param = (ParamType) interfaceParam;
            PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();

	    // map interfaceParam to a parameter of this class
            return "((" + adIntName + ")" + tr.whereObjExpr() + ")."
		+ ts.ParamSigBinding(param, curScope).name()
                + "()";
        } else {
            return "\"" + interfaceParam + "\"";
        }
    }

    /**
     *  Walks up the hierarchy of interfaces implemented by 
     *  <code>subtype</code>, adding all polymorphic interfaces 
     *  to superInterfaces.
     */
    protected static void fetchPolySuperInterfaces(Translator tr,
	ClassType subType, Map superInterfaces) 
    {
	    //BUG: subType should be allowed to be a Wrapped type in
	    //  addition to an Inst type.
        if (subType.flags().isInterface()) {
            PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
	    if (ts.isInstantiated(subType)) {
		ClassType u = ts.uninstantiate(subType);
                setForKey(superInterfaces, u).add(subType);
            }
	}

        Iterator directSuperInterfaces = subType.interfaces().listIterator();
        while (directSuperInterfaces.hasNext()) {
            ClassType parent = (ClassType) directSuperInterfaces.next();
            fetchPolySuperInterfaces(tr, parent, superInterfaces);
        }
    }

    /** 
     *  Gets the HashSet corresponding to key in map,
     *  creating it if not found.
     */
    protected static Set setForKey(Map map, Object key) {
        Set value = (Set) map.get(key);
        if (value == null) {
            value = new HashSet();
            map.put(key, value);
        }
        return value;
    }
}
