package polyglot.ext.polyj.types;

import polyglot.ast.*;
import polyglot.main.Report;
import polyglot.types.*;
import polyglot.util.*;

import polyglot.ext.jl.types.*;

import polyglot.ext.polyj.ast.*;
import polyglot.ext.polyj.visit.*;

import java.util.*;
import java.io.*;

/**
 * Instances of ParamType represent formal parameters of a
 * PolyJClassType.  Two ParamType instances are equal iff
 * they belong to the same PolyJClassType and are in the same
 * Position in its arguments list.
 */
public class ParamType_c extends ClassType_c implements ParamType {

    /**
     * The name of this parameter.
     */
    protected String name;

    /**
     * The position of this parameter in the parameter list of the constrained class.
     */
    protected int index;
    
    /**
     * The type this parameter constrains.
     */
    protected /*Type*/ PolyJClassType constrainedType;

    protected List fields;
    protected List methods;
    protected List constructors;
    protected Flags flags;

    public ParamType_c(TypeSystem ts, Position position, Type constrainedType, 
        String name, int index) 
    {
        super(ts, position);
        this.constrainedType = (PolyJClassType) constrainedType;
        this.name = name;
        this.index = index;
        
        this.fields = new TypedList(new LinkedList(), FieldInstance.class, false);
        this.methods = new TypedList(new LinkedList(), MethodInstance.class, false);
        this.constructors = new TypedList(new LinkedList(), ConstructorInstance.class, false);
        
        // Add the .default field of this parameter
        this.fields.add(ts.fieldInstance(position(), this,
            ts.Public().Static().Final(),
            this, "default"));
        
        this.flags = ts.Public().Static().Final();
    }

    public Kind kind() {
        return PARAM;
    }

    public List fields() {
        return Collections.unmodifiableList(fields);
    }
    
    public List methods() {
        return Collections.unmodifiableList(methods);
    }
    
    public List constructors() {
        return Collections.unmodifiableList(constructors);
    }
    
    public Flags flags() { return flags; }

    public Type constrainedType() {
        return constrainedType;
    }

    public PolyJClassType constrainedClass() {
	if (constrainedType instanceof PolyJClassType) {
	    return (PolyJClassType) constrainedType;
	} else {
	    return null;
	}
    }
    
    public int index() {
        return index;
    }

    public String toString() {
        return constrainedType() + "#" + fullName();
    }

    public String fullName() {
         return name;
    }

    public String name() {
        return name;
    }

    public void addConstraint(ClassConstraint c) {
        for (Iterator it = c.constraints().iterator(); it.hasNext(); ) {
            ProcedureConstraint pc = (ProcedureConstraint) it.next();
            if (pc instanceof MethodConstraint) {
                MethodConstraint mc = (MethodConstraint) pc;
                methods.add(mc.toInstance(this));
            } else if (pc instanceof ConstructorConstraint) {
                ConstructorConstraint cc = (ConstructorConstraint) pc;
                ConstructorInstance ci = (ConstructorInstance) cc.toInstance(this);
                if (Report.should_report(Report.types, 1)) {
    		    Report.report(1, "Adding to " + name + " CI: " + ci);
		}
                constructors.add(ci);
            } else {
                throw new InternalCompilerError("Illegal constraint.");
            }
        }
    }

    public polyglot.types.Package package_() { 
        //we're in the "package" of our polymorphic type
        return constrainedType.toClass().package_();
    }
    
    public Type superType() {
        return ts.Object(); //check me
    }

    public List interfaces() { return Collections.EMPTY_LIST; }

    public String translate(Resolver c) {
        return "java.lang.Object";
        //    return "##ParameterType#"+toString();
    }

    public boolean isNested() { return false; }
    public boolean isLocal() { return false; }
    public boolean isAnonymous() { return false; }
    public boolean inStaticContext() { return false; }
    public ClassType outer() { return null; }
    public List memberClasses() { return Collections.EMPTY_LIST; }
    public ClassType memberClassNamed(String s) { return null; }

    private void writeObject(java.io.ObjectOutputStream out) throws IOException
    {
	if (Report.should_report(Report.serialize, 3)) {
	    Report.report(3, "***Writing " + this);
	}
	out.defaultWriteObject();
	if (Report.should_report(Report.serialize, 3)) {
	    Report.report(3, "***Wrote " + this);
	}
    }
    
    private void readObject(java.io.ObjectInputStream in) 
	throws IOException, ClassNotFoundException 
    {
	if (Report.should_report(Report.serialize, 3)) {
	    Report.report(3, "***Reading ParamType");
	}
	in.defaultReadObject();
        if (Report.should_report(Report.serialize, 3)) {
	    Report.report(3, "***Read ParamType");
	}
    }

    public boolean isTopLevel() {
        // We have to override the super implementation because
        // ParamType's don't have kind TOP_LEVEL, but they aren't
        // nested or inner, either.  Perhaps ParamType's aren't really 
        // top-level types, but that's what they're most similar to.
        return true;
    }
}
